#!/usr/bin/env python2
#
# schlim.py   --  Experimental remote login manager
#
# (c) Copyright 2010 Michael Towers (larch42 at googlemail dot com)
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program; if not, write to the Free Software
# Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
#
#-------------------------------------------------------------------
# 2010.11.03

"""
Start with:
(exec) xinit /opt/schlim/schlim.py -- /usr/bin/X -nolisten tcp
"""

import os, sys, pwd, spwd, crypt, shutil, time
from subprocess import call, Popen, PIPE, STDOUT
import pexpect

import schlim_gui
import xmlrpclib

# Get configuration options
from schlim_conf import *
if os.path.isfile('/etc/schlim.conf'):
    execfile('/etc/schlim.conf')

# The configuration file can specify a user to be automatically logged in.
# As autologin should only happen the first time after start up (?)
# there is a flag, the file BLOCKAUTOLOGIN, to block it afterwards.

# Change to schlim directory
os.chdir(schlimdir)

import __builtin__
__builtin__.theme = theme

import locale
# Try to work around problems when the system encoding is not utf8
encoding = locale.getdefaultlocale()[1]


def log(message):
    sys.stderr.write(message.encode(encoding) + '\n')
    return message


def getlocalusers():
    return ([(p[0], p[4].split(',', 1)[0])
                for p in pwd.getpwall()
                        if (p[2] >= 500) and p[5].startswith('/home/') or p[5].startswith('/var/home/')]
                    + EXTRAUSERS)


class Schlim(schlim_gui.SchlimGui):
    def __init__(self):
        self.localuserlist = getlocalusers()
        lul = [u[0] for u in self.localuserlist]
        if autologin and (autologin in lul) and not os.path.isfile(BLOCKAUTOLOGIN):
            # Try to log this user in. If successful the call will not return.
            # Remote autologins are not supported (it seems a bad idea,
            # also the password is necessary, which I don't want to save).
            with open(BLOCKAUTOLOGIN, 'w') as fh:
                fh.write("Disable automatic login\n")
            self.login(autologin)

        self.username = ''
        self.hostname = Popen('hostname', stdout=PIPE).communicate()[0].strip()

        try:
            self.xmlrpc = xmlrpclib.ServerProxy(xmlrpc_server)
            rul = self.xmlrpc.getusers()
        except:
            rul = []
        self.remoteuserlist = []
        hidden = []
        for u in rul:
            try:
                pwdobj = pwd.getpwnam(u[0])
                hidden.append(u[0])
            except KeyError:
                self.remoteuserlist.append(u)

        ulist = self.localuserlist + self.remoteuserlist
        self.userlist = [('%s (%s)' % (u[0], u[1] if u[1]
                else _("Anonymous"))) for u in ulist]

        schlim_gui.SchlimGui.__init__(self)
        if hidden:
            self.message(_("Remote user(s) '%s' not available because of"
                    " local user with same name") % ', '.join(hidden))


    def cb_login(self, user, pw=None):
        """Called from the gui to perform one of the available actions,
        either logging in or shutting down.
        If the call is successful the gui will quit.
        Otherwise a message will be posted.
        """
        self.username = user
        if user in [u[0] for u in self.localuserlist]:
            # Local login
            em = self.local_authenticate(pw)
            if em != None:
                return em

        elif user in [u[0] for u in self.remoteuserlist]:
            # Remote login
            em = self.remote_setup(pw)
            if em != None:
                return em

        elif user in ('*restart*', '*shutdown*'):
            with open(SCHLIMEXITCODE, 'w') as fh:
                fh.write(user)
            return None

        else:
            return _("Invalid login name")

        return 'USER=' + user


    def local_authenticate(self, pw):
        try:
            pwcrypt = spwd.getspnam(self.username)[1]
        except:
            return _("Invalid local user: '%s'") % self.username
        if crypt.crypt(pw, pwcrypt) != pwcrypt:
            return _("Wrong password for '%s'") % self.username
        return None


    def remote_setup(self, pw):
        """Prepare for a remote login:
        - Check the share mount is clean
        - Mount the remote share (this is where the password is checked)
        - Fetch the necessary information about the user account
        - Add the remote user locally
        """
        if not os.path.isdir(HOMESDIR):
            os.makedirs(HOMESDIR)
        if os.listdir(HOMESDIR):
            return log(_("The remote home directory area is not clean,\n"
                    " please notify the system administrator"))

        cmd = (('sshfs -o allow_other -o default_permissions'
                ' -o StrictHostKeyChecking=no -o NumberOfPasswordPrompts=1'
                ' %s@%s:%s %s') % (self.username, SERVER, HOMESDIR, HOMESDIR))
        # As an alternative to '-o StrictHostKeyChecking=no',
        # /etc/ssh/ssh_known_hosts can contain the server info (and set
        # StrictHostKeyChecking=yes, but this is awkward when an unprepared
        # PC gets plugged in.
        # If the known hosts file is extended automatically it might be a
        # good idea to delete it automatically?
        # --- rm -f /root/.ssh/known_hosts
        output, rc = pexpect.run(cmd, withexitstatus=True,
                events={'(?i)password': pw+'\n'})
        if rc != 0:
            log(_("Remote share access failed:\n") + output)
            return _("Login failed - wrong password?")

        homedir = os.path.join(HOMESDIR, self.username)
        userinfo = self.xmlrpc.getuserinfo(self.username)
        groups = ','.join(userinfo[6])
        if groups != '':
            groups = '-G ' + groups
        cmd = 'useradd -c "%s" -g %d %s -M -d %s -u %d -s "%s" %s' % (
                userinfo[3], userinfo[2], groups,  homedir,
                userinfo[1], userinfo[5], self.username)
        if call(cmd, shell=True) != 0:
            return log(_("User account creation problem, please notify the"
                    " system administrator"))
        return None


    def login(self, user, display=':0'):
        try:
            pwdinfo = pwd.getpwnam(user)
        except:
            log(_("Invalid user: %s") % user)
            return None
        uid = pwdinfo[2]
        gid = pwdinfo[3]
        homedir = pwdinfo[5]
        shell = pwdinfo[6]

        if homedir.startswith('/var/home/'):
            tempdir = '/tmp/schlim_home_%s' % user
            if os.path.exists(tempdir):
                log(_("The temporary home directory for '%s' exists already,\n"
                        " please notify the system administrator"))
                sys.exit(1)
            else:
                if ((call(['cp', '-a', homedir, tempdir]) != 0)
                        or (call(['mount', '--bind', tempdir, homedir]) != 0)):
                    log(_("The temporary home directory for '%s' could not be prepared,\n"
                            " please notify the system administrator"))
                    sys.exit(1)

        xauth = homedir + '/.Xauthority'
#        call(['xauth', '-f', xauth, 'generate', display, '.', 'trusted'])
#        if os.path.isfile(xauth):
#            os.chown(xauth, uid, gid)

        xrootauth = '/root/.Xauthority'
        if os.path.isfile(xrootauth):
            call(['mv',  xrootauth, xauth])
            os.chown(xauth, uid, gid)

        os.execlp('su', 'su', '--login', user, schlimdir + '/xinitrc-user')


if __name__ == '__main__':
    ui = Schlim()
    rc = ui.start_gui()     # Enter the main event loop
    user = ui.username
    ui = None
    sys.exit(rc)

